package com.almworks.jira.structure.customitemtype;

import com.almworks.integers.LongArray;
import com.almworks.integers.LongList;
import com.almworks.jira.structure.api.auth.StructureAuth;
import com.almworks.jira.structure.api.forest.item.ItemForestBuilder;
import com.almworks.jira.structure.api.generator.GeneratorUnavailableException;
import com.almworks.jira.structure.api.generator.ItemChangeFilter;
import com.almworks.jira.structure.api.generator.UpdateChecker;
import com.almworks.jira.structure.api.generator.util.AbstractGenerator;
import com.almworks.jira.structure.api.util.StructureUtil;
import com.atlassian.jira.event.AbstractProjectEvent;
import com.atlassian.jira.permission.ProjectPermissions;
import com.atlassian.jira.project.Project;
import com.atlassian.jira.project.ProjectCategory;
import com.atlassian.jira.project.ProjectManager;
import com.atlassian.jira.security.PermissionManager;
import com.atlassian.jira.user.ApplicationUser;
import com.atlassian.jira.util.ErrorCollection;
import com.atlassian.jira.util.Supplier;
import org.jetbrains.annotations.NotNull;

import java.util.*;
import java.util.stream.Collectors;

/**
 * <p>This Inserter inserts {@link CustomProjectItemType custom project items}.
 * It inserts only projects those match the selected {@link #CATEGORY project category} parameter.</p>
 * <p>This class demonstrates how to:</p>
 * <ul>
 *   <li>Implement an Inserter</li>
 *   <li>{@link #CATEGORY Configure} the Insert with a web form and provide {@link #addDefaultFormParameters(Map) defaults}</li>
 *   <li>{@link #buildParametersFromForm(Map, ErrorCollection) Persist} parameters of the Inserter</li>
 *   <li>{@link #createFragment(Map, GenerationContext, ItemForestBuilder) Populate} a structure</li>
 * </ul>
 * A velocity template of the Inserter configuration form is defined in atlassian-plugin.xml file. See the 'resource' tag.
 */
public class CustomProjectInserter extends AbstractGenerator.Inserter {
  /**
   * <p>Name of a parameter of this inserter.</p>
   * <p>The parameter filters projects by project category.
   * The value is either a numeric project category {@link ProjectCategory#getId() ID} or one of predefined text values:
   * {@link #C_ALL}, {@link #C_NONE}.</p>
   * <p>This parameter name is used as a web form parameter and as a persisted configuration parameter.</p>
   */
  public static final String CATEGORY = "category";
  /**
   * <p>A special value of {@link #CATEGORY} parameter.
   * This value matches projects from of any category, including projects with no category.</p>
   * <p>This value is used only as web form value. The "all-projects" choice persists as <code>null</code> value</p>
   */
  private static final String C_ALL = "all";
  /**
   * <p>A special value of {@link #CATEGORY} parameter. This value matches projects with no category.</p>
   * <p>This value is used as both web form and persisted configuration parameter.</p>
   */
  private static final String C_NONE = "none";
  private final ProjectManager myProjectManager;
  private final PermissionManager myPermissionManager;

  public CustomProjectInserter(ProjectManager projectManager, PermissionManager permissionManager) {
    myProjectManager = projectManager;
    myPermissionManager = permissionManager;
  }

  @Override
  public void createFragment(@NotNull Map<String, Object> parameters, @NotNull GenerationContext context, @NotNull ItemForestBuilder builder) {
    String category = StructureUtil.getSingleParameter(parameters, CATEGORY);
    Supplier<Collection<Project>> projectSource;
    if (category == null) projectSource = myProjectManager::getProjects;
    else if (C_NONE.equals(category)) projectSource = myProjectManager::getProjectObjectsWithNoCategory;
    else projectSource =() -> {
        long categoryId;
        try {
          categoryId = Long.parseLong(category);
        } catch (NumberFormatException e) {
          return Collections.emptyList();
        }
        ProjectCategory projectCategory = myProjectManager.getProjectCategory(categoryId);
        return projectCategory != null ? myProjectManager.getProjectsFromProjectCategory(projectCategory) : Collections.emptyList();
      };
    projectSource = filterOutInaccessible(projectSource);
    ArrayList<Project> list = new ArrayList<>(projectSource.get());
    list.sort(Comparator.comparing(Project::getName));
    list.stream().map(CustomProjectItemType::createId).forEach(builder::nextRow);
    watchChanges(context, projectSource, list);
  }

  /**
   * Removes projects those the Structure owner is not permitted to browse.
   * See security note at {@link com.almworks.jira.structure.api.generator.StructureGenerator.Inserter#createFragment(Map, GenerationContext, ItemForestBuilder)}
   * @param projectSource original project set
   * @return a supplier of visible projects
   */
  private Supplier<Collection<Project>> filterOutInaccessible(Supplier<Collection<Project>> projectSource) {
    ApplicationUser user = StructureAuth.getUser();
    return () ->
      projectSource.get().stream()
        .filter(project -> myPermissionManager.hasPermission(ProjectPermissions.BROWSE_PROJECTS, project, user))
        .collect(Collectors.toList());
  }

  /**
   * <p>This method lets Structure know when JIRA projects changes and the generated forest fragment becomes stale.</p>
   * <p>It demonstrates two possible implementations. You may choose the one that best suits your needs.</p>
   * @param context the generation context
   * @param projectSource the Inserter's configuration
   * @param generated list of projects added to the forest fragment
   * @see #createItemChangeFilter()
   * @see #createUpdateChecker(Supplier, Collection)
   */
  private void watchChanges(@NotNull GenerationContext context, Supplier<Collection<Project>> projectSource, Collection<Project> generated) {
    // Only one of the following change detectors is required. UpdateChecker detector is commented out.
    // You may uncomment UpdateChecker detector and comment out ItemChangeFilter detector

//    context.addUpdateChecker(createUpdateChecker(projectSource, generated));
    context.addItemChangeFilter(createItemChangeFilter());
  }

  /**
   * Detects forest changes with {@link ItemChangeFilter}
   * <h3>Advantage over UpdateChecker</h3>
   * If no one project is changed this check quickly reports no changes. It need not to load current projects from a DB.
   * @return an implementation that returns true if any of changed items is a project.
   *         Such change may mean that the project has appeared or is removed.
   * @see CustomProjectItemType#onProjectEvent(AbstractProjectEvent)
   */
  @NotNull
  private ItemChangeFilter createItemChangeFilter() {
    return (itemIds, context) -> itemIds.stream().anyMatch(itemId ->
      CustomProjectItemType.KEY.equals(itemId.getItemType())
    );
  }

  /**
   * Detects forest changes with {@link UpdateChecker}
   * <h3>Advantages over ItemChangeFilter</h3>
   * <ul>
   *   <li>Does not require project changes to be reported to {@link com.almworks.jira.structure.api.item.ItemTracker}.</li>
   *   <li>Does not trigger forest regeneration when actual project change does not affect the forest.
   *   For example if a project name is changes there is no need to regenerate the forest.</li>
   *   <li>Detects project set change when current user permissions change.</li>
   * </ul>
   * @param projectSource the Inserter's configuration
   * @param generated list of projects added to the forest fragment
   * @return an implementation that returns true if current configured project set differs from the forest's content.
   */
  @NotNull
  private UpdateChecker createUpdateChecker(Supplier<Collection<Project>> projectSource, Collection<Project> generated) {
    LongList projectIds = collectProjectIds(generated);
    return () -> !projectIds.equals(collectProjectIds(projectSource.get()));
  }

  /**
   * @param projects a collection of JIRA projects
   * @return a sorted list of unique project IDs
   */
  private LongList collectProjectIds(Collection<Project> projects) {
    LongArray ids = new LongArray();
    projects.forEach(project -> ids.add(project.getId()));
    ids.sortUnique();
    return ids;
  }

  public void addDefaultFormParameters(@NotNull Map<String, Object> form) throws GeneratorUnavailableException {
    form.put(CATEGORY, C_ALL);
  }

  public void addParametersToForm(@NotNull Map<String, Object> parameters, @NotNull Map<String, Object> form) throws GeneratorUnavailableException {
    Object category = parameters.get(CATEGORY);
    form.put(CATEGORY, category != null ? category : C_ALL);
  }

  @NotNull
  public Map<String, Object> buildParametersFromForm(@NotNull Map<String, Object> form, @NotNull ErrorCollection errors) throws GeneratorUnavailableException {
    String category = StructureUtil.getSingleParameter(form, CATEGORY);
    if (C_ALL.equals(category)) category = null;
    return mapOf(CATEGORY, category);
  }

}
