package com.almworks.jira.structure.labelsextender;

import com.almworks.jira.structure.api.auth.StructureAuth;
import com.almworks.jira.structure.api.forest.item.ItemForestBuilder;
import com.almworks.jira.structure.api.generator.*;
import com.almworks.jira.structure.api.generator.util.*;
import com.almworks.jira.structure.api.item.CoreIdentities;
import com.almworks.jira.structure.api.item.ItemIdentity;
import com.almworks.jira.structure.api.row.StructureRow;
import com.almworks.jira.structure.api.util.StructureRunnable;
import com.almworks.jira.structure.api.util.StructureUtil;
import com.atlassian.jira.bc.issue.IssueService;
import com.atlassian.jira.bc.issue.label.LabelService;
import com.atlassian.jira.issue.Issue;
import com.atlassian.jira.issue.label.Label;
import com.atlassian.jira.user.ApplicationUser;
import com.atlassian.jira.util.*;
import com.google.common.collect.ImmutableMap;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.*;
import java.util.stream.Collectors;

/**
 * <p>This class is declared as a &lt;structure-extender&gt; module in atlassian-plugin.xml. It contains all the logic
 * necessary to implement the extender. The extender itself analyzes Labels for a given issue, and if there's a label
 * that is also an issue key, adds the referred issue under the analyzed issue.</p>
 *
 * <p>The methods in this class can be roughly separated into four categories:</p>
 *
 * <ol>
 *   <li>Managing generator's configuration.</li>
 *   <li>Generating forest.</li>
 *   <li>Processing actions coming from the user (moves, additions and removes).</li>
 *   <li>Utility methods.</li>
 * </ol>
 *
 * <p>The configuration consists of a single boolean value "sameProject". More complex generators may have more
 * parameters. It's also possible to not have parameters at all.</p>
 */
public class LabelsExtender extends AbstractGenerator.Extender {
  private static final String SAME_PROJECT_PARAMETER = "sameProject";

  private final LabelService myLabelService;
  private final IssueService myIssueService;
  private final I18nHelper myI18n;

  /**
   * Since this class is instantiated as a bean, you can get reference to services by injection. If you need
   * any Structure service, inject StructureComponents and get the services from there.
   */
  public LabelsExtender(LabelService labelService, IssueService issueService, I18nHelper i18n) {
    myLabelService = labelService;
    myIssueService = issueService;
    myI18n = i18n;
  }

  /**
   * This is the main generation function. It must achieve two things: a) provide an extension function, b)
   * provide a function that would let Structure understand when the generation should be redone.
   *
   * @param parameters stored generator parameters (see {@link #buildParametersFromForm(Map, ErrorCollection)}
   * @param context context, which provides additional information and functionality
   * @return a function that is used to generate child issues from parent issues
   */
  @Nullable
  @Override
  public ExtenderFunction getExtenderFunction(@NotNull Map<String, Object> parameters,
    @NotNull GenerationContext context)
  {
    // Retrieve options from stored parameters. The map is a deserialized JSON object, which was previously
    // constructed with buildParametersFromForm().
    boolean sameProject = StructureUtil.getSingleParameterBoolean(parameters, SAME_PROJECT_PARAMETER);

    // Structure tracks the updates to issues and offers generators a chance to indicate that a re-generation
    // is needed in response to some item changes. In our case, if a user changes some issue that the generator
    // has previously inspected, we want to check it again to see if the labels were changed.

    // For this to work, we need to call context.addItemChangeFilter() and provide it with a "filter" that
    // answers a question whether an item change is "interesting" for the generator. We're using a standard
    // BasicItemChangeFilter, which collects all the item IDs that the generator touched -- see the extender
    // function's code below.
    RecordingItemChangeFilter itemChangeFilter = BasicItemChangeFilter.createRecording();
    context.addItemChangeFilter(itemChangeFilter);

    // And this is the main function that does the generation.
    return new LabelsExtenderFunction(itemChangeFilter, sameProject);
  }


  /**
   * <p>This is the function that handles user actions. Actions are moves, additions and removals that the user
   * does to the forest. If the user does that with the dynamic content, the generator is given a chance to interpret
   * that action as the desire to change some information that is used to build the dynamic content.</p>
   *
   * <p>In our case, if the user moves generated issues between parents, we translate that into changing the labels
   * on the parent issues.</p>
   *
   * @param parameters stored generator parameters (see {@link #buildParametersFromForm(Map, ErrorCollection)}
   * @return action handler that is responsible for doing something in response to the user's action, or null if generator
   * can't respond to actions.
   */
  @Nullable
  @Override
  public ActionHandler.ExtenderActionHandler createActionHandler(@NotNull Map<String, Object> parameters) {
    boolean sameProject = StructureUtil.getSingleParameterBoolean(parameters, SAME_PROJECT_PARAMETER);
    return new LabelsExtenderActionHandler(sameProject);
  }

  /**
   * This function is called to populate form parameters map from the stored parameters map, before the form is shown.
   * The form parameters can then be used in the UI form -- see labelsextender-form.vm.
   *
   * @param parameters stored parameters
   * @param form writable map for form rendering parameters
   * @throws GeneratorUnavailableException if this generator is currently unavailable
   */
  @Override
  public void addParametersToForm(@NotNull Map<String, Object> parameters,
    @NotNull Map<String, Object> form) throws GeneratorUnavailableException
  {
    form.put(SAME_PROJECT_PARAMETER, StructureUtil.getSingleParameterBoolean(parameters, SAME_PROJECT_PARAMETER));
  }

  /**
   * This function is called to populate stored parameters map from the form data, after the form is submitted.
   * Whatever parameters your generator needs, extract them from form data and store in the returned map. Make
   * sure you store only simple objects that can be converted to JSON.
   *
   * @param form form data
   * @param errors if user input is invalid, add errors here and the form will be shown again
   * @return a map with generator parameters that will be stored as JSON
   * @throws GeneratorUnavailableException if this generator is currently unavailable
   */
  @NotNull
  @Override
  public Map<String, Object> buildParametersFromForm(@NotNull Map<String, Object> form,
    @NotNull ErrorCollection errors) throws GeneratorUnavailableException
  {
    return ImmutableMap.of(SAME_PROJECT_PARAMETER, StructureUtil.getSingleParameterBoolean(form, SAME_PROJECT_PARAMETER));
  }

  /**
   * This function is called to populate parameters map for rendering Summary (shown in the Structure grid when
   * Automation Editing Mode is on). See labelsextender-summary.vm.
   *
   * @param parameters stored parameters
   * @param summaryParameters writable map for form rendering parameters
   */
  @Override
  public void addParametersForSummary(@NotNull Map<String, Object> parameters,
    @NotNull Map<String, Object> summaryParameters)
  {
    summaryParameters.put(SAME_PROJECT_PARAMETER, StructureUtil.getSingleParameterBoolean(parameters, SAME_PROJECT_PARAMETER));
  }


  /**
   * This is the main function that does the generation. It is worth noting that the function is
   * going to be applied recursively. We only need to provide the extension based on the parent, and then
   * maybe this function will be called for all the children we have extended the structure with.
   */
  private class LabelsExtenderFunction implements ExtenderFunction {
    private final RecordingItemChangeFilter myItemChangeFilter;
    private final boolean mySameProject;

    public LabelsExtenderFunction(RecordingItemChangeFilter itemChangeFilter, boolean sameProject) {
      myItemChangeFilter = itemChangeFilter;
      mySameProject = sameProject;
    }

    /**
     * Helps optimize the process by answering a definitive "not interested" to all non-issue rows
     * and those issues that the user has no access to. This check must be very quick. It usually checks
     * item type.
     *
     * @param row the row that can be extended
     * @return true if this extender can possibly extend the row.
     */
    @Override
    public boolean isApplicableTo(StructureRow row) {
      return CoreIdentities.isIssue(row.getItemId()) && row.getItem(Issue.class) != null;
    }

    /**
     * Extends the structure. Given {@code row}, this function should use {@code builder} to build a sub-forest
     * that will be placed under that row.
     *
     * @param row parent row
     * @param builder recipient of the added content
     */
    @Override
    public void extend(@NotNull StructureRow row, @NotNull ItemForestBuilder builder) {
      Issue issue = row.getItem(Issue.class);
      if (issue == null) {
        // must not happen because of the way isApplicableTo() written, but we'll check anyway
        return;
      }

      // Record the fact that we've inspected this item. If it changes in the future, the generator will be called
      // again to do the extending once again.
      myItemChangeFilter.recordItem(row);

      // In our example, we analyze the labels and come up with the issues that we want to insert under the
      // parent row. In your case, this can be anything else.
      List<Issue> referredIssues = getReferredIssues(issue, mySameProject);

      // Use builder to create a sub-forest of issues.
      for (Issue referredIssue : referredIssues) {
        builder.nextRow(CoreIdentities.issue(referredIssue.getId()));
      }
    }

    /**
     * This is a utility function that we're using to extract the future child issues based on the parent issue.
     */
    private List<Issue> getReferredIssues(@NotNull Issue parentIssue, boolean sameProject) {
      // Use StructureAuth.getUser() for all authorization purposes.
      ApplicationUser currentUser = StructureAuth.getUser();

      LabelService.LabelsResult labels = myLabelService.getLabels(currentUser, parentIssue.getId());
      if (labels == null || !labels.isValid()) return Collections.emptyList();

      Long limitToProject = sameProject ? parentIssue.getProjectId() : null;

      return labels.getLabels().stream()
        .map(Label::getLabel) // String label
        .filter(JiraKeyUtils::validIssueKey) // String label, only valid issue keys
        .map(label -> myIssueService.getIssue(currentUser, label)) // IssueResult or null
        .filter(Objects::nonNull) // IssueResult
        .filter(IssueService.IssueResult::isValid) // IssueResult, only valid
        .map(IssueService.IssueResult::getIssue) // Issue referredIssue, assuming it can be null just in case
        .filter(Objects::nonNull) // Issue referredIssue
        .filter(issue -> !sameProject || Objects.equals(issue.getProjectId(), limitToProject)) // Issue referredIssue, filtered by project
        .collect(Collectors.toList());
    }
  }


  /**
   * <p>The action handler is responsible for doing something in response to the user's action -- change something in the
   * original data storage so that the next generator's run would create a structure that the user wants to see.</p>
   *
   * <p>Structure will call this extension handler only if it thinks that the generator can indeed respond to the user's action.</p>
   */
  private class LabelsExtenderActionHandler implements ActionHandler.ExtenderActionHandler {
    private final boolean mySameProject;

    public LabelsExtenderActionHandler(boolean sameProject) {
      mySameProject = sameProject;
    }

    /**
     * <p>This function is called whenever the user wants to change the parent of some row, and that falls within
     * this generator's sphere of influence.</p>
     *
     * <p>The action on the user interface could be move, addition, removal of rows. The only case where this function
     * is not called is when the user changes the order of children without changing the parent -- in that case,
     * {@link #reorderExtension} is called.</p>
     *
     * <p>The function uses {@code context} to react, and it can respond in several ways:</p>
     *
     * <ul>
     *   <li>React to the action by producing an <em>effect</em>.</li>
     *   <li><em>Block</em> the action with an explanatory text -- the user will see an error panel and the action will be reverted.</li>
     *   <li><em>Yield</em> to other generators, which may handle the action -- if others don't handle it, then it becomes a <em>block</em>.</li>
     * </ul>
     *
     * @param extension the row being moved, added or removed to the generator's scope
     * @param from the position where the extension row was before, or {@code null} if it was outside generator's scope
     * or not in structure
     * @param to the position where the user wants the extension row to be, or {@code null} if it is being moved
     * away from the generator's scope or removed from structure
     * @param context the context that provides additional information and ways to react
     */
    @Override
    public void moveExtension(@NotNull StructureRow extension, @Nullable StructurePosition from,
      @Nullable StructurePosition to, @NotNull HandlingContext context)
    {
      // See if this is an issue
      if (!CoreIdentities.isIssue(extension.getItemId())) {
        context.yield(myI18n.getText("sample.labelsextender.action.not-an-issue"));
        return;
      }

      // Get the Issue object and check if the user has access to it
      Issue movedIssue = extension.getItem(Issue.class);
      if (movedIssue == null) {
        context.block(myI18n.getText("sample.labelsextender.action.no-access"));
        return;
      }

      // So we want to add or remove a label, which would be equal to the moved issue's key.
      String changedLabel = movedIssue.getKey();

      if (changedLabel == null) {
        // This is a special case when an issue is being created under the generator's sphere of influence.
        // In that case, we get a chance to block the creation of issue at this point if we can check the
        // issue object and see that we wouldn't be able to place it at the desired position. In our case,
        // we could have checked that the intended project ID is the same as the parent's project ID and call yield()
        // if that is not true (in case we have "sameProject" setting on).
        //
        // The simple way though is to react with an empty effect. The action handler will be called again when the object
        // is created, and then we'll be able to set the label on the parent issue. However, if that second operation
        // fails, we'll get a new issue that is created but not added to the structure.
        context.effect(ctx -> {}, null);
        return;
      }

      // We'll use these variables to store the parameters for the Effect below. It's better not to store Issue and
      // other heavy-weight objects that can also get outdated, because the Effect can sit in memory for a while,
      // for undo/redo support.

      Long fromIssueId = null; // issue ID of the previous parent
      String fromIssueKey = null; // issue Key of the previous parent
      Long toIssueId = null; // issue ID of the future parent
      String toIssueKey = null; // issue Key of the future parent

      assert !(from == null && to == null);

      if (from != null) {
        // we're analyzing "under" coordinate -- a row that is the former parent of the moved row
        Issue fromIssue = getParentIssue(from.getUnder(), context);
        if (fromIssue == null) {
          // there was some error, context has been updated in getParentIssue()
          return;
        }
        fromIssueId = fromIssue.getId();
        fromIssueKey = fromIssue.getKey();
      }

      if (to != null) {
        // we're analyzing "under" coordinate -- a row that is the future parent of the moved row
        Issue toIssue = getParentIssue(to.getUnder(), context);
        if (toIssue == null) {
          // there was some error, context has been updated in getParentIssue()
          return;
        }
        // if we have "sameProject" turned on, make sure we don't put the issue under a different project's issue
        if (mySameProject && !Objects.equals(toIssue.getProjectId(), movedIssue.getProjectId())) {
          context.yield(myI18n.getText("sample.labelsextender.action.not-same-project"));
          return;
        }
        toIssueId = toIssue.getId();
        toIssueKey = toIssue.getKey();
      }

      // Provide the effect to Structure engine. Note that no actual change is executed in this function --
      // we just provide the effect and not execute the action itself.
      //
      // We also provide the reverse effect to support Undo/Redo. It's optional but highly recommended. The reverse
      // effect should do exactly the opposite of the effect, and it should be possible to use the forward effect again
      // after the reverse effect -- for Redo.
      ChangeLabelsEffect effect = new ChangeLabelsEffect(changedLabel, fromIssueId, fromIssueKey, toIssueId, toIssueKey);
      ChangeLabelsEffect undoEffect = new ChangeLabelsEffect(changedLabel, toIssueId, toIssueKey, fromIssueId, fromIssueKey);
      context.effect(effect, undoEffect);
    }

    @Nullable
    private Issue getParentIssue(@NotNull StructureRow row, HandlingContext context) {
      ItemIdentity itemId = row.getItemId();
      if (!CoreIdentities.isIssue(itemId)) {
        context.yield(myI18n.getText("sample.labelsextender.action.not-an-issue"));
        return null;
      }

      IssueService.IssueResult ir = myIssueService.getIssue(StructureAuth.getUser(), itemId.getLongId());
      if (!ir.isValid()) {
        // here we can analyze ir.getErrorCollection() and provide more descriptive error
        context.block(myI18n.getText("sample.labelsextender.action.no-access"));
        return null;
      }

      return ir.getIssue();
    }


    /**
     * This method is called if the user wants to rearrange the order in which the children are coming. In our case
     * we don't support reordering, so we just block the operation.
     *
     * @param extension the row being moved
     * @param from previous position (analyze "before/after" coordinates)
     * @param to future position, the "under" coordinate must be the same as in {@code from}
     * @param context the context that provides additional information and ways to react
     */
    @Override
    public void reorderExtension(@NotNull StructureRow extension, @NotNull StructurePosition from,
      @NotNull StructurePosition to, @NotNull HandlingContext context)
    {
      context.block(myI18n.getText("sample.labelsextender.action.reorder-blocked"));
    }

    /**
     * <p>This method provides the description of what's going to happen when an issue is "added"
     * to the generator's scope -- in our case, it means that we'll add a label to the parent issue.</p>
     *
     * <p>This label is used
     * when there are several ways to interpret a user's action (when there are several generators) and Structure has
     * to offer the user a choice, for example, "Create an issue link to the parent OR add a label to the parent?"</p>
     */
    @NotNull
    @Override
    public String getAddOptionDescription() {
      return myI18n.getText("sample.labelsextender.action.add.description");
    }

    /**
     * <p>Like {@link #getAddOptionDescription()}, describes the action that is going to take place when an issue
     * is moved to another parent.</p>
     */
    @NotNull
    @Override
    public String getMoveOptionDescription() {
      return myI18n.getText("sample.labelsextender.action.move.description");
    }

    /**
     * <p>Like {@link #getAddOptionDescription()}, describes the action that is going to take place when children
     * order is changed. We don't support that operation so we don't describe it.</p>
     */
    @NotNull
    @Override
    public String getReorderOptionDescription() {
      return "";
    }
  }


  /**
   * <p>The effect class encapsulates the effect of the user's action. It can be called repeatedly (with Undo/Redo), so
   * it must be pretty much independent, with parameters telling it what to do.</p>
   *
   * <p>An effect has to call one of the {@link EffectContext} methods to either provide a runnable code that executes
   * the action, or a reason for why the action cannot be executed.</p>
   *
   * <p>In our example, there are three possible effects: a label is removed from old parent, a label is added to
   * the new parent, and a combination of these two. Each effect has its own description, which is shown in a
   * pop-up message to the user after it has been successfully applied.</p>
   *
   * <p>Note that {@code changedLabel}, {@code fromIssueKey} and {@code toIssueKey} parameters are issue keys,
   * and issue keys may actually be changed for an issue if it's moved to another project. A more robust implementation
   * would just keep issue IDs as effect's fields, and then resolve issue ID to issue Key every time the action is
   * invoked.</p>
   */
  private class ChangeLabelsEffect implements ActionEffect {
    private final String myChangedLabel;
    private final Long myFromIssueId;
    private final String myFromIssueKey;
    private final Long myToIssueId;
    private final String myToIssueKey;

    public ChangeLabelsEffect(String changedLabel, Long fromIssueId, String fromIssueKey, Long toIssueId,
      String toIssueKey)
    {
      myChangedLabel = changedLabel;
      myFromIssueId = fromIssueId;
      myFromIssueKey = fromIssueKey;
      myToIssueId = toIssueId;
      myToIssueKey = toIssueKey;
    }

    @Override
    public void apply(EffectContext context) {
      ApplicationUser user = StructureAuth.getUser();
      StructureRunnable remove = null;
      StructureRunnable add = null;

      if (myFromIssueId != null) {
        // If we're removing a label from an issue, prepare the action itself -- possibly getting some errors.

        LabelService.LabelsResult labels = myLabelService.getLabels(user, myFromIssueId);
        if (!labels.isValid()) {
          // here we can analyze labels.getErrorCollection() and provide more descriptive error
          context.block(myI18n.getText("sample.labelsextender.action.error"));
          return;
        }
        Set<String> updatedLabels = labels.getLabels().stream()
          .map(Label::getLabel)
          .filter(label -> label != null && !label.equals(myChangedLabel))
          .collect(Collectors.toSet());

        LabelService.SetLabelValidationResult vr = myLabelService.validateSetLabels(user, myFromIssueId, updatedLabels);
        if (!vr.isValid()) {
          // here we can analyze vr.getErrorCollection() and provide more descriptive error
          context.block(myI18n.getText("sample.labelsextender.action.error"));
          return;
        }

        remove = () -> {
          myLabelService.setLabels(user, vr, true, true);
        };
      }

      if (myToIssueId != null) {
        // If we're adding a label to an issue, also prepare.

        LabelService.AddLabelValidationResult vr = myLabelService.validateAddLabel(user, myToIssueId, myChangedLabel);
        if (!vr.isValid()) {
          // here we can analyze vr.getErrorCollection() and provide more descriptive error
          context.block(myI18n.getText("sample.labelsextender.action.error"));
          return;
        }

        add = () -> {
          myLabelService.addLabel(user, vr, true);
        };
      }

      // Finally, pass the StructureRunnable effect to the context, which will then execute it.
      // The String message appears shortly in a notification pop-up to the user.

      if (add != null) {
        if (remove != null) {
          // remove + add at the same time
          context.effect(
            myI18n.getText("sample.labelsextender.action.removed-and-added-label", myChangedLabel, myFromIssueKey, myToIssueKey),
            remove.andThen(add));

        } else {
          // just add
          context.effect(
            myI18n.getText("sample.labelsextender.action.added-label", myChangedLabel, myToIssueKey),
            add);
        }
      } else if (remove != null) {
        // just remove
        context.effect(
          myI18n.getText("sample.labelsextender.action.removed-label", myChangedLabel, myFromIssueKey),
          remove);
      }
    }
  }
}
