package com.almworks.jira.structure.scheduledsync;

import com.almworks.integers.LongIterator;
import com.almworks.integers.LongList;
import com.almworks.jira.structure.api.StructureComponents;
import com.almworks.jira.structure.api.auth.StructureAuth;
import com.almworks.jira.structure.api.job.ScheduledJob;
import com.almworks.jira.structure.api.job.ScheduledJobManager;
import com.almworks.jira.structure.api.process.*;
import com.almworks.jira.structure.api.sync.*;
import com.atlassian.activeobjects.external.ActiveObjects;
import com.atlassian.cache.CacheManager;
import com.atlassian.event.api.EventListener;
import com.atlassian.event.api.EventPublisher;
import com.atlassian.fugue.Option;
import com.atlassian.plugin.event.events.PluginEnabledEvent;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.DisposableBean;

/**
 * @author Igor Sereda
 * @author Igor Baltiyskiy
 */
public class Scheduler implements DisposableBean {
  private static final Logger logger = LoggerFactory.getLogger(Scheduler.class);

  private final Schedule mySchedule;
  private final ScheduledJobManager myJobManager;
  private final StructureSyncManager mySyncManager;
  private final ProcessHandleManager myProcessHandleManager;
  
  private final EventPublisher myEventPublisher;

  public Scheduler(StructureComponents services, ActiveObjects activeObjects, CacheManager cacheManager, EventPublisher eventPublisher) {
    mySchedule = new Schedule(activeObjects, cacheManager);
    myJobManager = services.getScheduledJobManager();
    mySyncManager = services.getSyncManager();
    myProcessHandleManager = services.getProcessHandleManager();
    myEventPublisher = eventPublisher;
    myEventPublisher.register(this);
  }

  @EventListener
  public void onPluginEnabled(PluginEnabledEvent event) {
    if (Schedule.PLUGIN_KEY.equals(event.getPlugin().getKey())) {
      start();
    }
  }

  private void start() {
    LongList syncIds = mySchedule.getAllScheduledSyncs();
    logger.warn(this + " starts schedules for " + syncIds);
    for (LongIterator syncIdIt : syncIds) {
      long syncId = syncIdIt.value();
      myJobManager.addJob(getJobPropertyKey(syncId), new ScheduledSyncJob(syncId));
    }
  }

  public void destroy() throws Exception {
    myEventPublisher.unregister(this);
    
    LongList syncIds = mySchedule.getAllScheduledSyncs();
    logger.warn(this + " stops schedules for " + syncIds);
    for (LongIterator syncId : syncIds) {
      myJobManager.removeJob(getJobPropertyKey(syncId.value()));
    }
  }

  public String toString() {
    return "SyncScheduler";
  }

  public Schedule.ScheduledSync getSchedule(long syncId) {
    return mySchedule.getSchedule(syncId).getOrNull();
  }

  public void setSchedule(long syncId, long period) {
    if (mySchedule.setSchedule(syncId, period)) {
      String jobPropertyKey = getJobPropertyKey(syncId);
      if (period > 0) {
        myJobManager.addJob(jobPropertyKey, new ScheduledSyncJob(syncId));
      } else {
        myJobManager.removeJob(jobPropertyKey);
      }
    }
  }

  public long getLastRunTime(@Nullable Schedule.ScheduledSync schedule) {
    return schedule == null ? 0 : myJobManager.getLastRunTime(getJobPropertyKey(schedule.getSyncId())); 
  }

  public long getNextRunTime(@Nullable Schedule.ScheduledSync schedule, long lastRunTime, long now) {
    if (schedule == null) return 0;
    if (lastRunTime != 0) {
      return lastRunTime + schedule.getPeriod();
    } else {
      // Have just enabled schedule, will run as soon as possible
      return now;
    }
  }
  
  private static String getJobPropertyKey(long syncId) {
    return Schedule.PROPERTY_PREFIX + ".sync." + syncId;
  }


  private class ScheduledSyncJob implements ScheduledJob {
    private final long mySyncId;

    private ScheduledSyncJob(long syncId) {
      mySyncId = syncId;
    }

    @Override
    public boolean shouldRun(long lastRunTime) {
      long now = System.currentTimeMillis();
      Option<Schedule.ScheduledSync> schedule = getSchedule();
      if (!schedule.isDefined()) {
        return false;
      }
      long period = schedule.get().getPeriod();
      return now - lastRunTime >= period;
    }

    @Override
    public void run() throws Exception {
      Option<Schedule.ScheduledSync> schedule = getSchedule();
      if (!schedule.isDefined()) return;

      SyncInstance instance = mySyncManager.getInstalledSynchronizer(mySyncId);
      if (instance == null) {
        logger.warn(this + " cannot run " + schedule + ": no synchronizer");
        return;
      }

      ProcessDisplayParameters processDisplayParams =
        new ProcessDisplayParameters.Builder().withProcessName("Scheduled resync for synchronizer #" + mySyncId)
          .build();
      ProcessHandle process = myProcessHandleManager.createNew(StructureAuth.getUser(), processDisplayParams);
      mySchedule.setResyncProcess(mySyncId, process);
      try {
        mySyncManager.resync(mySyncId, false, process.getId());
      } catch (StructureSynchronizerException e) {
        logger.warn(this + " cannot run " + schedule, e);
      }
    }

    @NotNull
    private Option<Schedule.ScheduledSync> getSchedule() {
      Option<Schedule.ScheduledSync> schedule = mySchedule.getSchedule(mySyncId);
      if (!schedule.isDefined()) {
        // JDC note: The schedule might have been removed on another node. Here's when we learn that on other nodes.
        myJobManager.removeJob(getJobPropertyKey(mySyncId));
      }
      return schedule;
    }
  }  
}
