package com.almworks.jira.structure.sbcolumn;

import com.almworks.jira.structure.api.attribute.CoreAttributeSpecs;
import com.almworks.jira.structure.api.error.StructureProviderException;
import com.almworks.jira.structure.api.export.*;
import com.almworks.jira.structure.api.util.StructureUtil;
import com.almworks.jira.structure.api.view.ViewSpecification;
import com.atlassian.jira.config.StatusManager;
import com.atlassian.jira.issue.status.Status;
import org.apache.commons.lang.StringUtils;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.*;

/**
 * Simple text-based renderer provider for the Status Bar column.
 * Used when no specialized provider is available.
 */
public class StatusBarRendererProvider implements ExportRendererProvider {
  static final String COLUMN_KEY = "com.almworks.jira.structure.sbcolumn";

  private final StatusManager myStatusManager;

  public StatusBarRendererProvider(StatusManager statusManager) {
    myStatusManager = statusManager;
  }

  @Nullable
  @Override
  public ExportRenderer getColumnRenderer(@NotNull ExportFormat exportFormat, @NotNull ViewSpecification.Column columnSpec, @NotNull ExportRequestContext context) throws StructureProviderException {
    if (!COLUMN_KEY.equals(columnSpec.getKey())) {
      return null;
    }

    String name = columnSpec.getName();
    if (StringUtils.isEmpty(name)) {
      name = context.getI18n().getText("sbcolumn.name");
    }

    Map<String, Object> parameters = columnSpec.getParameters();
    List<String> statuses = StructureUtil.getMultiParameter(parameters, "statuses");
    if (statuses.isEmpty()) {
      throw new StructureProviderException("statuses parameter is required");
    }
    boolean includeItself = StructureUtil.getSingleParameterBoolean(parameters, "includeItself");

    return new StatusBarRenderer(name, statuses, includeItself);
  }

  private class StatusBarRenderer implements ExportRenderer {
    private final String myName;
    private final List<String> myStatuses;
    private final boolean myIncludeItself;

    private final Map<String, String> myStatusNames = new HashMap<String, String>();

    public StatusBarRenderer(String name, List<String> statuses, boolean includeItself) {
      myName = name;
      myStatuses = statuses;
      myIncludeItself = includeItself;
    }

    @Override
    public void prepare(@NotNull ExportRequestContext context) {
      context.requireAttribute(StatusBarAttributeProvider.STATUS_BAR);
      context.requireAttribute(CoreAttributeSpecs.STATUS);
    }

    @Override
    public void configureColumn(@NotNull ExportColumn column, @NotNull ExportRenderContext context) {
      column.getHeaderCell().setText(myName);
    }

    @Override
    public void renderCell(@NotNull ExportCell cell, @NotNull ExportRow row, @NotNull ExportRenderContext context) {
      Map<String, Integer> map = row.get(StatusBarAttributeProvider.STATUS_BAR);
      if (map == null || map.isEmpty()) {
        return;
      }

      String itself = null;
      if (!myIncludeItself) {
        if (row.isLeaf()) {
          return;
        }
        Status status = row.get(CoreAttributeSpecs.STATUS);
        if (status != null) {
          itself = status.getId();
        }
      }

      StringBuilder sb = new StringBuilder();
      for (String id : myStatuses) {
        if (StringUtils.isEmpty(id)) {
          continue;
        }
        Integer count = map.get(id);
        if (count == null) {
          continue;
        }
        String name = getStatusName(id);
        if (StringUtils.isEmpty(name)) {
          continue;
        }
        if (id.equals(itself)) {
          count--;
        }
        if (count <= 0) {
          continue;
        }
        if (sb.length() > 0) {
          sb.append(", ");
        }
        sb.append(count).append(' ').append(name);
      }

      if (sb.length() > 0) {
        cell.setText(sb.toString());
      }
    }

    private String getStatusName(String id) {
      String name = myStatusNames.get(id);
      if (name == null) {
        Status status = myStatusManager.getStatus(id);
        if (status != null) {
          name = status.getName();
          myStatusNames.put(id, name);
        }
      }
      return name;
    }
  }
}
